@@ -1,19 +1,12 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
-from .animate import animate
-from .colors import red, bold
-from .emojies import stars, hazard, sleep
-from .util import rmdir, mkdir, PipxError, WINDOWS
-from .Venv import Venv
-from pathlib import Path
-from shutil import which
-from typing import List, Optional
 import datetime
 import distutils.spawn
 import hashlib
 import http.client
 import logging
+import multiprocessing
 import os
 import shlex
 import shutil
@@ -21,9 +14,28 @@ import subprocess
 import textwrap
 import time
 import urllib.parse
-from .constants import LOCAL_BIN_DIR, PIPX_PACKAGE_NAME, PIPX_VENV_CACHEDIR
+from pathlib import Path
+from shutil import which
+from typing import List, Optional
 
-TEMP_VENV_EXPIRATION_THRESHOLD_DAYS = 2
+from .animate import animate
+from .colors import bold, red
+from .constants import (
+    LOCAL_BIN_DIR,
+    PIPX_PACKAGE_NAME,
+    PIPX_VENV_CACHEDIR,
+    TEMP_VENV_EXPIRATION_THRESHOLD_DAYS,
+)
+from .emojies import hazard, sleep, stars
+from .util import (
+    WINDOWS,
+    PipxError,
+    get_pypackage_bin_path,
+    mkdir,
+    rmdir,
+    run_pypackage_bin,
+)
+from .Venv import Venv
 
 
 def run(
@@ -33,6 +45,7 @@ def run(
     python: str,
     pip_args: List[str],
     venv_args: List[str],
+    pypackages: bool,
     verbose: bool,
     use_cache: bool,
 ):
@@ -65,6 +78,19 @@ def run(
         binary = f"{binary}.exe"
         logging.warning(f"Assuming binary is {binary!r} (Windows only)")
 
+    pypackage_bin_path = get_pypackage_bin_path(binary)
+    if pypackage_bin_path.exists():
+        logging.info(
+            f"Using binary in local __pypackages__ directory at {str(pypackage_bin_path)}"
+        )
+        return run_pypackage_bin(pypackage_bin_path, binary_args)
+    if pypackages:
+        raise PipxError(
+            f"'--pypackages' flag was passed, but {str(pypackage_bin_path)!r} was "
+            "not found. See https://github.com/cs01/pythonloc to learn how to "
+            "install here, or omit the flag."
+        )
+
     venv_dir = _get_temporary_venv_path(package_or_url, python, pip_args, venv_args)
 
     venv = Venv(venv_dir)
@@ -73,10 +99,10 @@ def run(
 
     if bin_path.exists():
         logging.info(f"Reusing cached venv {venv_dir}")
-        return venv.run_binary(binary, binary_args)
+        retval = venv.run_binary(binary, binary_args)
     else:
         logging.info(f"venv location is {venv_dir}")
-        return _download_and_run(
+        retval = _download_and_run(
             Path(venv_dir),
             package_or_url,
             binary,
@@ -87,6 +113,10 @@ def run(
             verbose,
         )
 
+    if not use_cache:
+        rmdir(venv_dir)
+    return retval
+
 
 def _download_and_run(
     venv_dir: Path,
@@ -102,10 +132,10 @@ def _download_and_run(
     venv.create_venv(venv_args, pip_args)
     venv.install_package(package, pip_args)
     if not (venv.bin_path / binary).exists():
-        binaries = venv.get_package_binary_paths(package)
+        binaries = venv.get_venv_metadata_for_package(package).binaries
         raise PipxError(
             f"{binary} not found in package {package}. Available binaries: "
-            f"{', '.join(b.name for b in binaries)}"
+            f"{', '.join(b for b in binaries)}"
         )
     return venv.run_binary(binary, binary_args)
 
@@ -146,7 +176,7 @@ def _prepare_venv_cache(venv: Venv, bin_path: Path, use_cache: bool):
 def _remove_all_expired_venvs():
     for venv_dir in Path(PIPX_VENV_CACHEDIR).iterdir():
         if _is_temporary_venv_expired(venv_dir):
-            logging.infof(f"Removing expired venv {str(venv_dir)}")
+            logging.info(f"Removing expired venv {str(venv_dir)}")
             rmdir(venv_dir)
 
 
@@ -169,7 +199,10 @@ def upgrade(
     verbose: bool,
     *,
     upgrading_all: bool,
+    include_deps: bool,
 ) -> int:
+    """Returns nonzero if package was upgraded, 0 if version did not change"""
+
     if not venv_dir.is_dir():
         raise PipxError(
             f"Package is not installed. Expected to find {str(venv_dir)}, "
@@ -178,7 +211,7 @@ def upgrade(
 
     venv = Venv(venv_dir, verbose=verbose)
 
-    old_version = venv.get_package_version(package)
+    old_version = venv.get_venv_metadata_for_package(package).package_version
     do_animation = not verbose
     try:
         with animate(f"upgrading pip for package {package_or_url!r}", do_animation):
@@ -189,7 +222,14 @@ def upgrade(
 
     with animate(f"upgrading package {package_or_url!r}", do_animation):
         venv.upgrade_package(package_or_url, pip_args)
-    new_version = venv.get_package_version(package)
+    new_version = venv.get_venv_metadata_for_package(package).package_version
+
+    metadata = venv.get_venv_metadata_for_package(package)
+    _expose_binaries_globally(LOCAL_BIN_DIR, metadata.binary_paths, package)
+
+    if include_deps:
+        for _, binary_paths in metadata.binary_paths_of_dependencies.items():
+            _expose_binaries_globally(LOCAL_BIN_DIR, binary_paths, package)
 
     if old_version == new_version:
         if upgrading_all:
@@ -199,17 +239,16 @@ def upgrade(
                 f"{package} is already at latest version {old_version} (location: {str(venv_dir)})"
             )
         return 0
-
-    binary_paths = venv.get_package_binary_paths(package)
-    _expose_binaries_globally(LOCAL_BIN_DIR, binary_paths, package)
-
-    print(
-        f"upgraded package {package} from {old_version} to {new_version} (location: {str(venv_dir)})"
-    )
-    return 1
+    else:
+        print(
+            f"upgraded package {package} from {old_version} to {new_version} (location: {str(venv_dir)})"
+        )
+        return 1
 
 
-def upgrade_all(pipx_local_venvs: Path, pip_args: List[str], verbose: bool):
+def upgrade_all(
+    pipx_local_venvs: Path, pip_args: List[str], verbose: bool, *, include_deps: bool
+):
     packages_upgraded = 0
     num_packages = 0
     for venv_dir in pipx_local_venvs.iterdir():
@@ -219,9 +258,18 @@ def upgrade_all(pipx_local_venvs: Path, pip_args: List[str], verbose: bool):
             package_or_url = PIPX_PACKAGE_NAME
         else:
             package_or_url = package
-        packages_upgraded += upgrade(
-            venv_dir, package, package_or_url, pip_args, verbose, upgrading_all=True
-        )
+        try:
+            packages_upgraded += upgrade(
+                venv_dir,
+                package,
+                package_or_url,
+                pip_args,
+                verbose,
+                upgrading_all=True,
+                include_deps=include_deps,
+            )
+        except Exception:
+            logging.error(f"Error encountered when upgrading {package}")
 
     if packages_upgraded == 0:
         print(
@@ -240,8 +288,14 @@ def install(
     verbose: bool,
     *,
     force: bool,
+    include_deps: bool,
 ):
-    if venv_dir.exists() and next(venv_dir.iterdir()):
+    try:
+        exists = venv_dir.exists() and next(venv_dir.iterdir())
+    except StopIteration:
+        exists = False
+
+    if exists:
         if force:
             print(f"Installing to existing directory {str(venv_dir)!r}")
         else:
@@ -258,46 +312,88 @@ def install(
         venv.remove_venv()
         raise
 
-    if venv.get_package_version(package) is None:
+    if venv.get_venv_metadata_for_package(package).package_version is None:
         venv.remove_venv()
         raise PipxError(f"Could not find package {package}. Is the name correct?")
 
-    binary_paths = venv.get_package_binary_paths(package)
-    if not binary_paths:
-        for dependent_package in venv.get_package_dependencies(package):
-            dependent_binaries = venv.get_package_binary_paths(dependent_package)
-            if dependent_binaries:
-                print(
-                    f"Note: Dependent package '{dependent_package}' contains {len(dependent_binaries)} binaries"
-                )
-            for b in dependent_binaries:
-                print(f"  - {b.name}")
+    _run_post_install_actions(venv, package, local_bin_dir, venv_dir, include_deps)
+
+
+def _run_post_install_actions(
+    venv: Venv, package: str, local_bin_dir: Path, venv_dir: Path, include_deps: bool
+):
+    metadata = venv.get_venv_metadata_for_package(package)
+    if not include_deps and not metadata.binary_paths:
+        for dep, dependent_binaries in metadata.binary_paths_of_dependencies.items():
+            print(
+                f"Note: Dependent package '{dep}' contains {len(dependent_binaries)} binaries"
+            )
+            for binary in dependent_binaries:
+                print(f"  - {binary.name}")
         venv.remove_venv()
-        raise PipxError(f"No binaries associated with package {package}.")
+        if len(metadata.binary_paths_of_dependencies.keys()):
+            raise PipxError(
+                f"No binaries associated with package {package}. "
+                "Try again with '--include-deps' to include binaries of dependent packages."
+            )
+        else:
+            raise PipxError(f"No binaries associated with package {package}. ")
 
-    _expose_binaries_globally(local_bin_dir, binary_paths, package)
-    _list_installed_package(venv_dir, new_install=True)
+    _expose_binaries_globally(local_bin_dir, metadata.binary_paths, package)
+    if include_deps:
+        for _, binary_paths in metadata.binary_paths_of_dependencies.items():
+            _expose_binaries_globally(local_bin_dir, binary_paths, package)
 
-    if not distutils.spawn.find_executable(str(binary_paths[0])):
+    print(_get_package_summary(venv_dir, package=package, new_install=True))
+
+    random_binary_name: str
+    if metadata.binaries:
+        random_binary_name = metadata.binaries[0]
+    else:
+        random_binary_name = metadata.binaries_of_dependencies[0]
+
+    _warn_if_not_on_path(local_bin_dir, random_binary_name)
+    print(f"done! {stars}")
+
+
+def _warn_if_not_on_path(local_bin_dir: Path, binary: str):
+    if not distutils.spawn.find_executable(binary):
         logging.warning(
             f"{hazard}  Note: {str(local_bin_dir)!r} is not on your PATH environment "
             "variable. These binaries will not be globally accessible until "
             "your PATH is updated. Run `pipx ensurepath` to automatically add it, "
             "or manually modify your PATH in your shell's config file (i.e. ~/.bashrc)."
         )
-    print(f"done! {stars}")
 
 
-def inject(venv_dir: Path, package: str, pip_args: List[str], verbose: bool):
+def inject(
+    venv_dir: Path,
+    package: str,
+    pip_args: List[str],
+    *,
+    verbose: bool,
+    include_binaries: bool,
+    include_deps: bool,
+):
     if not venv_dir.exists() or not next(venv_dir.iterdir()):
-        raise PipxError(f"Can't inject {package} into nonexistent venv {venv_dir}!")
+        raise PipxError(
+            textwrap.dedent(
+                f"""\
+            Can't inject {package!r} into nonexistent Virtual Environment {str(venv_dir)!r}.
+            Be sure to install the package first with pipx install {venv_dir.name!r}
+            before injecting into it."""
+            )
+        )
 
     venv = Venv(venv_dir, verbose=verbose)
     venv.install_package(package, pip_args)
 
-    if venv.get_package_version(package) is None:
+    if venv.get_venv_metadata_for_package(package).package_version is None:
         raise PipxError(f"Could not find package {package}. Is the name correct?")
 
+    if include_binaries:
+        _run_post_install_actions(venv, package, LOCAL_BIN_DIR, venv_dir, include_deps)
+
     print(f"done! {stars}")
 
 
@@ -313,15 +409,18 @@ def uninstall(venv_dir: Path, package: str, local_bin_dir: Path, verbose: bool):
 
     venv = Venv(venv_dir, verbose=verbose)
 
-    package_binary_paths = venv.get_package_binary_paths(package)
+    metadata = venv.get_venv_metadata_for_package(package)
+    binary_paths = metadata.binary_paths
+    for dep_paths in metadata.binary_paths_of_dependencies.values():
+        binary_paths += dep_paths
     for file in local_bin_dir.iterdir():
         if WINDOWS:
-            for b in package_binary_paths:
+            for b in binary_paths:
                 if file.name == b.name:
                     file.unlink()
         else:
             symlink = file
-            for b in package_binary_paths:
+            for b in binary_paths:
                 if symlink.exists() and b.exists() and symlink.samefile(b):
                     logging.info(f"removing symlink {str(symlink)}")
                     symlink.unlink()
@@ -343,6 +442,7 @@ def reinstall_all(
     pip_args: List[str],
     venv_args: List[str],
     verbose: bool,
+    include_deps: bool,
 ):
     for venv_dir in pipx_local_venvs.iterdir():
         package = venv_dir.name
@@ -359,6 +459,7 @@ def reinstall_all(
             venv_args,
             verbose,
             force=True,
+            include_deps=include_deps,
         )
 
 
@@ -398,53 +499,78 @@ def _symlink_package_binaries(
                 pass
             else:
                 logging.warning(
-                    f"{hazard}  File exists at {str(symlink_path)} and points to {symlink_path.resolve()}. Not creating."
+                    f"{hazard}  File exists at {str(symlink_path)} and points "
+                    f"to {symlink_path.resolve()}. Not modifying."
                 )
         else:
             shadow = which(binary)
-            symlink_path.symlink_to(b)
-            pass
+            try:
+                symlink_path.symlink_to(b)
+            except FileExistsError:
+                pass
+
             if shadow:
                 logging.warning(
                     f"{hazard}  Note: {binary} was already on your PATH at " f"{shadow}"
                 )
 
 
-def _list_installed_package(path: Path, *, new_install: bool = False) -> None:
+def _get_package_summary(
+    path: Path, *, package: str = None, new_install: bool = False
+) -> str:
     venv = Venv(path)
     python_path = venv.python_path.resolve()
-    package = path.name
+    if package is None:
+        package = path.name
+    metadata = venv.get_venv_metadata_for_package(package)
 
-    version = venv.get_package_version(package)
-    if version is None:
-        print(f"{package} is not installed in the venv {str(path)}")
-        return
-
-    python_version = venv.get_python_version()
-    package_binary_paths = venv.get_package_binary_paths(package)
-    package_binary_names = [b.name for b in package_binary_paths]
+    if metadata.package_version is None:
+        not_installed = red("is not installed")
+        return f"   package {bold(package)} {not_installed} in the venv {str(path)}"
 
+    binaries = metadata.binaries + metadata.binaries_of_dependencies
     exposed_binary_paths = _get_exposed_binary_paths_for_package(
-        venv.bin_path, package_binary_paths, LOCAL_BIN_DIR
+        venv.bin_path, binaries, LOCAL_BIN_DIR
     )
     exposed_binary_names = sorted(p.name for p in exposed_binary_paths)
     unavailable_binary_names = sorted(
-        set(package_binary_names) - set(exposed_binary_names)
+        set(metadata.binaries) - set(exposed_binary_names)
+    )
+    return _get_list_output(
+        metadata.python_version,
+        python_path,
+        metadata.package_version,
+        package,
+        new_install,
+        exposed_binary_names,
+        unavailable_binary_names,
     )
 
-    print(
-        f"  {'installed' if new_install else ''} package {bold(shlex.quote(package))} {bold(version)}, {python_version}"
+
+def _get_list_output(
+    python_version: str,
+    python_path: Path,
+    package_version: str,
+    package: str,
+    new_install: bool,
+    exposed_binary_names: List[str],
+    unavailable_binary_names: List[str],
+) -> str:
+    output = []
+    output.append(
+        f"  {'installed' if new_install else ''} package {bold(shlex.quote(package))} {bold(package_version)}, {python_version}"
     )
 
     if not python_path.exists():
-        logging.error(f"    associated python path {str(python_path)} does not exist!")
+        output.append(f"    associated python path {str(python_path)} does not exist!")
 
     if new_install and exposed_binary_names:
-        print("  These binaries are now globally available")
+        output.append("  These binaries are now globally available")
     for name in exposed_binary_names:
-        print(f"    - {name}")
+        output.append(f"    - {name}")
     for name in unavailable_binary_names:
-        print(f"    - {red(name)} (symlink not installed)")
+        output.append(f"    - {red(name)} (symlink not installed)")
+    return "\n".join(output)
 
 
 def list_packages(pipx_local_venvs: Path):
@@ -454,31 +580,31 @@ def list_packages(pipx_local_venvs: Path):
         return
 
     print(f"venvs are in {bold(str(pipx_local_venvs))}")
-    print(f"symlinks to binaries are in {bold(str(LOCAL_BIN_DIR))}")
-    for d in dirs:
-        _list_installed_package(d)
+    print(f"binaries are exposed on your $PATH at {bold(str(LOCAL_BIN_DIR))}")
+
+    with multiprocessing.Pool() as p:
+        for package_summary in p.map(_get_package_summary, dirs):
+            print(package_summary)
 
 
 def _get_exposed_binary_paths_for_package(
-    bin_path: Path, package_binary_paths: List[Path], local_bin_dir: Path
+    bin_path: Path, package_binary_names: List[str], local_bin_dir: Path
 ):
     bin_symlinks = set()
-    package_binary_names = {p.name for p in package_binary_paths}
     for b in local_bin_dir.iterdir():
         try:
             # sometimes symlinks can resolve to a file of a different name
             # (in the case of ansible for example) so checking the resolved paths
             # is not a reliable way to determine if the symlink exists.
             # windows doesn't use symlinks, so the check is less strict.
+            if WINDOWS and b.name in package_binary_names:
+                is_same_file = True
+            else:
+                is_same_file = b.resolve().parent.samefile(bin_path)
 
-            if b.name in package_binary_names:
-                if WINDOWS:
-                    is_same_file = True
-                else:
-                    is_same_file = b.resolve().parent.samefile(bin_path)
+            if is_same_file:
+                bin_symlinks.add(b)
 
-                if is_same_file:
-                    bin_symlinks.add(b)
         except FileNotFoundError:
             pass
     return bin_symlinks
@@ -501,11 +627,11 @@ def ensurepath(bin_dir: Path):
 
     if config_file and os.path.exists(config_file):
         with open(config_file, "a") as f:
-            f.write("\n# added by pipx (https://github.com/cs01/pipx)\n")
+            f.write("\n# added by pipx (https://github.com/pipxproject/pipx)\n")
             if "fish" in shell:
-                f.write("set -x PATH %s $PATH\n\n" % bin_dir)
+                f.write(f"set -x PATH {str(bin_dir)} $PATH\n\n")
             else:
-                f.write('export PATH="%s:$PATH"\n' % bin_dir)
+                f.write(f'export PATH="{str(bin_dir)}{os.pathsep}$PATH"\n')
         print(f"Added {str(bin_dir)} to the PATH environment variable in {config_file}")
         print("")
         print(f"Open a new terminal to use pipx {stars}")
